"""Distance measures to compare two probability density functions."""

import numpy as np

__version__ = '0.9.0'

EPSILON = np.finfo(float).eps

def acc(u, v):
    """The average of Manhattan and Chebyshev distances.

    Synonyms: 
        ACC distance
        Average distance

    References:
        1. Krause EF (2012) Taxicab Geometry An Adventure in Non-Euclidean 
           Geometry. Dover Publications.
        2. Sung-Hyuk C (2007) Comprehensive Survey on Distance/Similarity 
           Measures between Probability Density Functions. International
           Journal of Mathematical Models and Methods in Applied Sciences.
           vol. 1(4), pp. 300-307.
    """
    return (manhattan(u, v) + chebyshev(u, v))/2


def add_chisq(u, v):
    """Additive Symmetric Chi-square distance.

    References:
        1. Sung-Hyuk C (2007) Comprehensive Survey on Distance/Similarity 
           Measures between Probability Density Functions. International
           Journal of Mathematical Models and Methods in Applied Sciences.
           vol. 1(4), pp. 300-307.
    """
    uvmult = u * v
    with np.errstate(divide='ignore', invalid="ignore"):
        return np.sum(np.where(uvmult != 0, ((u-v)**2 * (u+v))/uvmult, 0))     


def bhattacharyya(u, v):
    """Bhattacharyya distance.

    Returns a distance value between 0 and 1.

    References:
        1. Bhattacharyya A (1947) On a measure of divergence between two 
           statistical populations defined by probability distributions, 
           Bull. Calcutta Math. Soc., 35, 99–109.
        2. Sung-Hyuk C. (2007) Comprehensive Survey on Distance/Similarity 
           Measures between Probability Density Functions. International
           Journal of Mathematical Models and Methods in Applied Sciences.
           1(4), 300-307.
        3. https://en.wikipedia.org/wiki/Bhattacharyya_distance
    """
    return -np.log(np.sum(np.sqrt(u*v)))


def braycurtis(u, v):
    """Bray-Curtis distance.

    Synonyms:
        Sørensen distance
        Bray-Curtis dissimilarity
    
    Notes:
        When used for comparing two probability density functions (pdfs),
        Bray-Curtis distance equals Manhattan distance divided by 2.

    References:
        1. Bray JR, Curtis JT (1957) An ordination of the upland forest of
           the southern Winsconsin. Ecological Monographies, 27, 325-349.
        2. Sung-Hyuk C. (2007) Comprehensive Survey on Distance/Similarity 
           Measures between Probability Density Functions. International
           Journal of Mathematical Models and Methods in Applied Sciences.
           1(4), 300-307.
        3. https://en.wikipedia.org/wiki/Bray–Curtis_dissimilarity
    """
    return np.sum(np.abs(u - v)) / np.sum(np.abs(u + v))


def canberra(u, v):
    """Canberra distance.

    Notes:
        When `u[i]` and `v[i]` are 0 for given i, then the fraction 0/0 = 0
        is used in the calculation.

    References:
        1. Sung-Hyuk C. (2007) Comprehensive Survey on Distance/Similarity 
           Measures between Probability Density Functions. International
           Journal of Mathematical Models and Methods in Applied Sciences.
           1(4), 300-307.
    """
    with np.errstate(invalid='ignore'):
        return np.nansum(np.abs(u - v) / (np.abs(u) + np.abs(v)))


def chebyshev(u, v):
    """Chebyshev distance.

    Synonyms:
        Chessboard distance
        King-move metric
        Maximum value distance
        Minimax approximation

    References:
        1. Sung-Hyuk C. (2007) Comprehensive Survey on Distance/Similarity 
           Measures between Probability Density Functions. International
           Journal of Mathematical Models and Methods in Applied Sciences.
           1(4), 300-307.        
    """
    return np.amax(np.abs(u - v))


def chebyshev_min(u, v):
    """Minimum value distance (my measure)."""
    return np.amin(np.abs(u - v))


def clark(u, v):
    """Clark distance.

    Clark distance equals the squared root of half of the divergence.

    Notes:
        When `u[i]` and `v[i]` are 0 for given i, then the fraction 0/0 = 0
        is used in the calculation.

    References:
        1. Sung-Hyuk C. (2007) Comprehensive Survey on Distance/Similarity 
           Measures between Probability Density Functions. International
           Journal of Mathematical Models and Methods in Applied Sciences.
           1(4), 300-307.
    """
    with np.errstate(divide='ignore', invalid="ignore"):
        return np.sqrt(np.nansum(np.power(np.abs(u-v)/(u+v),2)))


def cosine(u, v):
    """Cosine distance.

    References:
        1. SciPy.
    """
    return 1 - np.dot(u, v)/(np.sqrt(np.dot(u, u))*np.sqrt(np.dot(v, v)))


def correlation_pearson(u, v):
    """Pearson correlation distance.

    Returns a distance value between 0 and 2.
    """
    r = np.ma.corrcoef(u, v)[0, 1]
    return 1.0 - r


def czekanowski(u, v):
    """Czekanowski distance.

    References:
        1. Sung-Hyuk C. (2007) Comprehensive Survey on Distance/Similarity 
           Measures between Probability Density Functions. International
           Journal of Mathematical Models and Methods in Applied Sciences.
           1(4), 300-307.
    """
    return np.sum(np.abs(u - v)) / np.sum(u + v)


def dice(u, v):
    """Dice dissimilarity.

    Synonyms:
        Sorensen distance

    Referemces:
        1. Dice LR (1945) Measures of the amount of ecologic association
           between species. Ecology. 26, 297-302.
        2. Sung-Hyuk C. (2007) Comprehensive Survey on Distance/Similarity 
           Measures between Probability Density Functions. International
           Journal of Mathematical Models and Methods in Applied Sciences.
           1(4), 300-307.
    """
    u_v = u - v
    return np.dot(u_v, u_v) / (np.dot(u, u) + np.dot(v, v))


def divergence(u, v):
    """Divergence.

    Divergence equals squared Clark distance multiplied by 2.

    References:
        1. Sung-Hyuk C. (2007) Comprehensive Survey on Distance/Similarity 
           Measures between Probability Density Functions. International
           Journal of Mathematical Models and Methods in Applied Sciences.
           1(4), 300-307.
    """
    with np.errstate(invalid="ignore"):
        return 2 * np.nansum(np.power(u-v,2) / np.power(u+v,2))


def euclidean(u, v):
    """Euclidean distance.

    Synonyms:
        Pythagorean metric

    References:
        1. Sung-Hyuk C. (2007) Comprehensive Survey on Distance/Similarity 
           Measures between Probability Density Functions. International
           Journal of Mathematical Models and Methods in Applied Sciences.
           1(4), 300-307.        
    """
    return np.linalg.norm(u-v)


def google(u, v):
    """Normalized Google Distance (NGD).

    Returns a distance value between 0 and 1. Two sequences are treated
    as two different web pages and the each word frequency represents 
    terms found in each webpage.

    Notes:
        When used for comparing two probability density functions (pdfs),
        Google distance equals half of Manhattan distance.

    References:
        1. Lee & Rashid (2008) Information Technology, ITSim 2008.
           doi:10.1109/ITSIM.2008.4631601.
    """
    x = float(np.sum(u))
    y = float(np.sum(v))
    summin = float(np.sum(np.minimum(u, v)))
    return (max([x, y]) - summin) / ((x + y) - min([x, y]))


def gower(u, v):
    """Gower distance.

    Gower distance equals Manhattan distance divided by vector length.

    References:
        1. Gower JC. (1971) General Coefficient of Similarity
           and Some of Its Properties, Biometrics 27, 857-874.
        2. Sung-Hyuk C. (2007) Comprehensive Survey on Distance/Similarity 
           Measures between Probability Density Functions. International
           Journal of Mathematical Models and Methods in Applied Sciences.
           1(4), 300-307.
    """
    return np.sum(np.abs(u - v)) / u.size


def hellinger(u, v):
    """Hellinger distance.

    Notes:
        This implementation produces values two times larger than values
        obtained by Hellinger distance described in Wikipedia and also
        in https://gist.github.com/larsmans/3116927.

        Wikipedia:
        np.sqrt(np.sum((np.sqrt(u) - np.sqrt(v)) ** 2)) / np.sqrt(2) 

    References:
       1.  Sung-Hyuk C. (2007) Comprehensive Survey on Distance/Similarity 
           Measures between Probability Density Functions. International
           Journal of Mathematical Models and Methods in Applied Sciences.
           1(4), 300-307.
    """
    return np.sqrt(2*np.sum((np.sqrt(u) - np.sqrt(v)) ** 2))


def jaccard(u, v):
    """Jaccard distance.

    References:
       1.  Sung-Hyuk C. (2007) Comprehensive Survey on Distance/Similarity 
           Measures between Probability Density Functions. International
           Journal of Mathematical Models and Methods in Applied Sciences.
           1(4), 300-307.
    """
    uv = np.dot(u, v)
    return 1 - (uv / (np.dot(u, u) + np.dot(v, v) - uv))


def jeffreys(u, v, epsilon=EPSILON):
    """Jeffreys divergence.

    Synonyms:
        J divergence

    References:
        1. Jeffreys H (1946) An Invariant Form for the Prior Probability
           in Estimation Problems. Proc.Roy.Soc.Lon., Ser. A 186, 453-461.
        2. Sung-Hyuk C. (2007) Comprehensive Survey on Distance/Similarity 
           Measures between Probability Density Functions. International
           Journal of Mathematical Models and Methods in Applied Sciences.
           1(4), 300-307.
    """
    # Add epsilon to zeros in vectors to avoid division
    # by 0 and/or log of 0. Alternatively, zeros in the
    # vectors could be ignored or masked (see below).
    # u = ma.masked_where(u == 0, u)
    # v = ma.masked_where(v == 0, u)
    u = np.where(u==0, epsilon, u)
    v = np.where(v==0, epsilon, v)
    return np.sum((u-v) * np.log(u / v))


def jensenshannon_divergence(u, v, epsilon=EPSILON):
    """Jensen-Shannon divergence.

    Notes:
        1. Equals half of Topsøe distance
        2. Equals squared jensenshannon_distance.

    References:
        1. Lin J. (1991) Divergence measures based on the Shannon entropy.
           IEEE Transactions on Information Theory, 37(1):145–151.
        2. Sung-Hyuk C. (2007) Comprehensive Survey on Distance/Similarity 
           Measures between Probability Density Functions. International
           Journal of Mathematical Models and Methods in Applied Sciences.
           1(4), 300-307.

    Comments:
        Equals Jensen difference in Sung-Hyuk (2007):
        u = np.where(u==0, epsilon, u)
        v = np.where(v==0, epsilon, v)
        el1 = (u * np.log(u) + v * np.log(v)) / 2
        el2 = (u + v)/2
        el3 = np.log(el2)
        return np.sum(el1 - el2 * el3)
    """
    u = np.where(u==0, epsilon, u)
    v = np.where(v==0, epsilon, v)     
    dl = u * np.log(2*u/(u+v))
    dr = v * np.log(2*v/(u+v))
    return (np.sum(dl) + np.sum(dr)) / 2


def jensen_difference(u, v, epsilon=EPSILON):
    """Jensen difference

    Comments:
        Seems equal to Jensen-Shannon divergence.

    References:
        1. Sung-Hyuk C. (2007) Comprehensive Survey on Distance/Similarity 
           Measures between Probability Density Functions. International
           Journal of Mathematical Models and Methods in Applied Sciences.
           1(4), 300-307.
    """
    u = np.where(u==0, epsilon, u)
    v = np.where(v==0, epsilon, v)        
    el1 = (u * np.log(u) + v * np.log(v)) / 2
    el2 = (u + v) / 2
    return np.sum(el1 - el2 * np.log(el2))


def k_divergence(u, v, epsilon=EPSILON):
    """K divergence.

    References:
        1. Sung-Hyuk C. (2007) Comprehensive Survey on Distance/Similarity 
           Measures between Probability Density Functions. International
           Journal of Mathematical Models and Methods in Applied Sciences.
           1(4), 300-307.
    """
    u = np.where(u==0, epsilon, u)
    v = np.where(v==0, epsilon, v)
    return np.sum(u*np.log(2*u/(u+v)))


def kl_divergence(u, v, epsilon=EPSILON):
    """Kullback-Leibler divergence.

    Syonymes:
        KL divergence, relative entropy, information deviation

    References:
        1. Kullback S, Leibler RA (1951) On information and sufficiency.
           Ann. Math. Statist. 22:79–86
        2. Sung-Hyuk C. (2007) Comprehensive Survey on Distance/Similarity 
           Measures between Probability Density Functions. International
           Journal of Mathematical Models and Methods in Applied Sciences.
           1(4):300-307.
    """
    u = np.where(u==0, epsilon, u)
    v = np.where(v==0, epsilon, v)
    return np.sum(u * np.log(u / v))


def kulczynski(u, v):
    """Kulczynski distance.

    References:
        1. Sung-Hyuk C. (2007) Comprehensive Survey on Distance/Similarity 
           Measures between Probability Density Functions. International
           Journal of Mathematical Models and Methods in Applied Sciences.
           1(4):300-307.
    """
    return np.sum(np.abs(u - v)) / np.sum(np.minimum(u, v))


def kumarjohnson(u, v):
    """Kumar-Johnson distance.

    References:
        1. Kumar P, Johnson A. (2005) On a symmetric divergence measure
           and information inequalities, Journal of Inequalities in pure
           and applied Mathematics. 6(3).
        2. Sung-Hyuk C. (2007) Comprehensive Survey on Distance/Similarity 
           Measures between Probability Density Functions. International
           Journal of Mathematical Models and Methods in Applied Sciences.
           1(4):300-307.
    """
    uvmult = u*v
    with np.errstate(divide='ignore', invalid='ignore'):
        numer = np.power(u**2 - v**2, 2)
        denom = 2 * np.power(uvmult, 3/2)
        return np.sum(np.where(uvmult != 0, numer/denom, 0))


def lorentzian(u, v):
    """Lorentzian distance.

    References:
        1. Sung-Hyuk C. (2007) Comprehensive Survey on Distance/Similarity 
           Measures between Probability Density Functions. International
           Journal of Mathematical Models and Methods in Applied Sciences.
           1(4):300-307.        

    Notes:
        One (1) is added to guarantee the non-negativity property and to 
        eschew the log of zero
    """
    return np.sum(np.log(np.abs(u-v)+1))


def manhattan(u, v):
    """Manhattan distance.

    Synonyms:
        City block distance
        Rectilinear distance
        Taxicab norm

    Notes:
        Manhattan distance between two probability density functions 
        (pdfs) equals:
        1. Non-intersection distance multiplied by 2.
        2. Gower distance multiplied by vector length.
        3. Bray-Curtis distance multiplied by 2.
        4. Google distance multiplied by 2.

    References:
        1. Sung-Hyuk C. (2007) Comprehensive Survey on Distance/Similarity 
           Measures between Probability Density Functions. International
           Journal of Mathematical Models and Methods in Applied Sciences.
           1(4):300-307.
    """
    return np.sum(np.abs(u - v))


def marylandbridge(u, v):
    """Maryland Bridge distance.

    References:
        1. Deza M, Deza E (2009) Encyclopedia of Distances. 
           Springer-Verlag Berlin Heidelberg. 1-590.
    """
    uvdot = np.dot(u, v)
    return 1 - (uvdot/np.dot(u, u) + uvdot/np.dot(v, v))/2


def matusita(u, v):
    """Matusita distance.

    Notes:
        Equals square root of Squared-chord distance.

    References:
        1. Sung-Hyuk C. (2007) Comprehensive Survey on Distance/Similarity 
           Measures between Probability Density Functions. International
           Journal of Mathematical Models and Methods in Applied Sciences.
           1(4):300-307.
    """
    return np.sqrt(np.sum((np.sqrt(u)-np.sqrt(v))**2))


def max_symmetric_chisq(u, v):
    """Max-symmetric chisq.

    References:
        1. Sung-Hyuk C. (2007) Comprehensive Survey on Distance/Similarity 
           Measures between Probability Density Functions. International
           Journal of Mathematical Models and Methods in Applied Sciences.
           1(4):300-307.
    """
    return max(neyman_chisq(u, v), pearson_chisq(u, v))


def minkowski(u, v, p=2):
    """Minkowski distance.

    Parameters:
        p : int
            The order of the norm of the difference.

    Notes:
        When p goes to infinite, the Chebyshev distance is derived.

    References:
        1. Sung-Hyuk C. (2007) Comprehensive Survey on Distance/Similarity 
           Measures between Probability Density Functions. International
           Journal of Mathematical Models and Methods in Applied Sciences.
           1(4):300-307.
    """
    return np.linalg.norm(u - v, ord=p)


def motyka(u, v):
    """Motyka distance.

    Notes:
        The distance between identical vectors is not equal to 0 but 0.5.
    
    References:
        1. Sung-Hyuk C. (2007) Comprehensive Survey on Distance/Similarity 
           Measures between Probability Density Functions. International
           Journal of Mathematical Models and Methods in Applied Sciences.
           1(4), 300-307.
    """
    return np.sum(np.maximum(u, v)) / np.sum(u + v)


def neyman_chisq(u, v):
    """Neyman chi-square distance.

    References:
        1. Neyman J (1949) Contributions to the theory of the chi^2 test. 
           In Proceedings of the First Berkley Symposium on Mathematical
           Statistics and Probability.
        2. Sung-Hyuk C. (2007) Comprehensive Survey on Distance/Similarity 
           Measures between Probability Density Functions. International
           Journal of Mathematical Models and Methods in Applied Sciences.
           1(4), 300-307.
    """
    with np.errstate(divide='ignore', invalid='ignore'):
       return np.sum(np.where(u != 0, (u-v)**2/u, 0))


def nonintersection(u, v):
    """Distance based on intersection.

    Synonyms:
        Non-overlaps
        Intersection distance

    Notes:
        When used for comparing two probability density functions (pdfs),
        Nonintersection distance equals half of Manhattan distance.

    References:
        1. Sung-Hyuk C. (2007) Comprehensive Survey on Distance/Similarity 
           Measures between Probability Density Functions. International
           Journal of Mathematical Models and Methods in Applied Sciences.
           1(4), 300-307.
    """
    return 1 - np.sum(np.minimum(u, v))


def pearson_chisq(u, v):
    """Pearson chi-square divergence.

    Notes:
        Pearson chi-square divergence is asymmetric.

    References:
        1. Pearson K. (1900) On the Criterion that a given system of 
           deviations from the probable in the case of correlated system
           of variables is such that it can be reasonable supposed to have
           arisen from random sampling, Phil. Mag. 50, 157-172.
        2. Sung-Hyuk C. (2007) Comprehensive Survey on Distance/Similarity 
           Measures between Probability Density Functions. International
           Journal of Mathematical Models and Methods in Applied Sciences.
           1(4), 300-307.
    """
    with np.errstate(divide='ignore', invalid='ignore'):
        return np.sum(np.where(v != 0, (u-v)**2/v, 0))


def penroseshape(u, v):
    """Penrose shape distance.

    References:
        1. Deza M, Deza E (2009) Encyclopedia of Distances. 
           Springer-Verlag Berlin Heidelberg. 1-590.
    """
    umu = np.mean(u)
    vmu = np.mean(v)
    return np.sqrt(np.sum(((u-umu)-(v-vmu))**2))


def soergel(u, v):
    """Soergel distance.

    Notes:
        Equals Tanimoto distance.

    References:
        1. Sung-Hyuk C. (2007) Comprehensive Survey on Distance/Similarity 
           Measures between Probability Density Functions. International
           Journal of Mathematical Models and Methods in Applied Sciences.
           1(4), 300-307.
    """
    return np.sum(np.abs(u - v)) / np.sum(np.maximum(u, v))


def squared_chisq(u, v):
    """Squared chi-square distance.

    Synonyms:
        Triangular discrimination

    References:
        1. Sung-Hyuk C. (2007) Comprehensive Survey on Distance/Similarity 
           Measures between Probability Density Functions. International
           Journal of Mathematical Models and Methods in Applied Sciences.
           1(4), 300-307.
    """
    uvsum = u + v
    with np.errstate(divide='ignore', invalid='ignore'):
        return np.sum(np.where(uvsum != 0, (u-v)**2/uvsum, 0))


def squaredchord(u, v):
    """Squared-chord distance.

    Notes:
        Equals to squared Matusita distance.

    Reference:
        1. Gavin DG et al. (2003) A statistical approach to evaluating 
           distance metrics and analog assignments for pollen records.
           Quaternary Research 60:356–367.
        2. Sung-Hyuk C. (2007) Comprehensive Survey on Distance/Similarity 
           Measures between Probability Density Functions. International
           Journal of Mathematical Models and Methods in Applied Sciences.
           1(4), 300-307.
    """
    return np.sum((np.sqrt(u) - np.sqrt(v))**2)


def squared_euclidean(u, v):
    """Squared Euclidean distance.

    Notes:
        Equals to squared Euclidean distance.

    Reference:
        1. Gavin DG et al. (2003) A statistical approach to evaluating 
           distance metrics and analog assignments for pollen records.
           Quaternary Research 60:356–367.
        2. Sung-Hyuk C. (2007) Comprehensive Survey on Distance/Similarity 
           Measures between Probability Density Functions. International
           Journal of Mathematical Models and Methods in Applied Sciences.
           1(4), 300-307.
    """
    return np.dot((u - v), (u - v))


def taneja(u, v, epsilon=EPSILON):
    """Taneja distance.

    References:
        1. Taneja IJ. (1995), New Developments in Generalized Information
           Measures, Chapter in: Advances in Imaging and Electron Physics,
           Ed. P.W. Hawkes, 91, 37-135.
        2. Sung-Hyuk C. (2007) Comprehensive Survey on Distance/Similarity 
           Measures between Probability Density Functions. International
           Journal of Mathematical Models and Methods in Applied Sciences.
           1(4), 300-307.
    """
    u = np.where(u==0, epsilon, u)
    v = np.where(v==0, epsilon, v)
    uvsum = u + v
    return np.sum((uvsum/2)*np.log(uvsum/(2*np.sqrt(u*v))))


def tanimoto(u, v):
    """Tanimoto distance.

    Notes:
        Equals Soergel distance.

    References:
        1. Sung-Hyuk C. (2007) Comprehensive Survey on Distance/Similarity 
           Measures between Probability Density Functions. International
           Journal of Mathematical Models and Methods in Applied Sciences.
           1(4), 300-307
    """
    #return np.sum(abs(u-v)) / np.sum(np.maximum(u, v))
    usum = np.sum(u)
    vsum = np.sum(v)
    minsum = np.sum(np.minimum(u, v))
    return (usum + vsum - 2*minsum) / (usum + vsum - minsum)


def topsoe(u, v, epsilon=EPSILON):
    """Topsøe distance.

    Synonyms:
        Information statistic

    Notes:
        Equals two times Jensen-Shannon divergence.

    References:
        1. Sung-Hyuk C (2007) Comprehensive Survey on Distance/Similarity 
           Measures between Probability Density Functions. International
           Journal of Mathematical Models and Methods in Applied Sciences.
           1(4), 300-307
    """
    u = np.where(u==0, epsilon, u)
    v = np.where(v==0, epsilon, v)        
    dl = u * np.log(2*u/(u+v))
    dr = v * np.log(2*v/(u+v))
    return np.sum(dl + dr)


def vicis_symmetric_chisq(u, v):
    """Vicis Symmetric chi-square distance.

    References:
        1. Sung-Hyuk C (2007) Comprehensive Survey on Distance/Similarity 
           Measures between Probability Density Functions. International
           Journal of Mathematical Models and Methods in Applied Sciences.
           1(4), 300-307
    """
    with np.errstate(divide='ignore', invalid='ignore'):
        u_v = (u - v)**2
        uvmin = np.minimum(u, v)**2
        return np.sum(np.where(uvmin != 0, u_v/uvmin, 0))


def vicis_wave_hedges(u, v):
    """Vicis-Wave Hedges distance.

    References:
        1. Sung-Hyuk C (2007) Comprehensive Survey on Distance/Similarity 
           Measures between Probability Density Functions. International
           Journal of Mathematical Models and Methods in Applied Sciences.
           1(4), 300-307
    """
    with np.errstate(divide='ignore', invalid='ignore'):
        u_v = abs(u - v)
        uvmin = np.minimum(u, v)
        return np.sum(np.where(uvmin != 0, u_v/uvmin, 0))


def wave_hedges(u, v):
    """Wave Hedges distance.

    References:
        1. Sung-Hyuk C (2007) Comprehensive Survey on Distance/Similarity 
           Measures between Probability Density Functions. International
           Journal of Mathematical Models and Methods in Applied Sciences.
           1(4), 300-307
    """
    with np.errstate(divide='ignore', invalid='ignore'):
        u_v = abs(u - v)
        uvmax = np.maximum(u, v)
        return np.sum(np.where(((u_v != 0) & (uvmax != 0)), u_v/uvmax, 0))



import numpy as np
import unittest

import distance

# Test case 1: non-zero values in u and v.
uv1 = [
   np.array([0.33, 0.21, 0.46]),
   np.array([0.32, 0.50, 0.18])
]

# Test case 2: zero value in v.
uv2 = [
   np.array([0.41, 0.23, 0.36]),
   np.array([0.30, 0.70, 0])
]

# Test case 3: zero value in u.
uv3 = [
   np.array([0.33, 0.67, 0]),
   np.array([0.50, 0.25, 0.25])
]

# Test case 4: zero value in u and v.
uv4 = [
   np.array([0.45, 0.55, 0]),
   np.array([0.68, 0.32, 0])
]

# Test case 5: u and v are identical.
uv5 = [
   np.array([0.20, 0.05, 0.40, 0.30, 0.05]),
   np.array([0.20, 0.05, 0.40, 0.30, 0.05])
]


class TestDistance(unittest.TestCase):

    def setUp(self):
        self.vectors = [uv1, uv2, uv3, uv4, uv5]

    def func_test(self, func, correct_values):
        for i, (u, v) in enumerate(self.vectors):
            self.assertAlmostEqual( 
                func(u, v), correct_values[i], places=6
            )

    def test_add_chisq(self):
        correct_values = [1.175282, 1.345852, 1.114259, 0.456844, 0]
        self.func_test(distance.add_chisq, correct_values)

    def test_acc(self):
        correct_values = [0.435, 0.705, 0.63, 0.345, 0]
        self.func_test(distance.acc, correct_values)

    def test_bhattacharyya(self):
        correct_values = [0.065340, 0.285070, 0.203991, 0.027683, 0]
        self.func_test(distance.bhattacharyya, correct_values)

    def test_braycurtis(self):
        correct_values = [0.29, 0.47, 0.42, 0.23, 0]
        self.func_test(distance.braycurtis, correct_values)

    def test_canberra(self):
        correct_values = [0.861335, 1.660306, 1.661341, 0.467908, 0]
        self.func_test(distance.canberra, correct_values)

    def test_chebyshev(self):
        correct_values = [0.29, 0.47, 0.42, 0.23, 0]
        self.func_test(distance.chebyshev, correct_values)

    def test_chebyshev_min(self):
        correct_values = [0.01, 0.11, 0.17, 0, 0]
        self.func_test(distance.chebyshev_min, correct_values)

    def test_clark(self):
        correct_values = [0.598728, 1.131109, 1.118196, 0.333645, 0]
        self.func_test(distance.clark, correct_values)

    def test_cosine(self):
        correct_values = [0.216689, 0.370206, 0.272996, 0.097486, 0]
        self.func_test(distance.cosine, correct_values)

    def test_correlation_pearson(self):
        correct_values = [1.995478, 1.755929, 1.008617, 0.254193, 0]
        self.func_test(distance.correlation_pearson, correct_values)

    def test_czekanowski(self):
        correct_values = [0.29, 0.47, 0.42, 0.23, 0]
        self.func_test(distance.czekanowski, correct_values)

    def test_dice(self):
        correct_values = [0.216974, 0.389641, 0.287093, 0.098897, 0]
        self.func_test(distance.dice, correct_values)

    def test_divergence(self):
        correct_values = [0.7169498, 2.5588168, 2.5007261, 0.222638, 0]
        self.func_test(distance.divergence, correct_values)

    def test_euclidean(self):
        correct_values = [0.403237, 0.602163, 0.517494, 0.325269, 0]
        self.func_test(distance.euclidean, correct_values)

    def test_google(self):
        correct_values = [0.29, 0.47, 0.42, 0.23, 0]
        self.func_test(distance.google, correct_values)

    def test_gower(self):
        correct_values = [0.193333, 0.313333, 0.28, 0.153333, 0]
        self.func_test(distance.gower, correct_values) 

    def test_hellinger(self):
        correct_values = [0.5029972, 0.996069, 0.859140, 0.330477, 0]
        self.func_test(distance.hellinger, correct_values) 

    def test_jaccard(self):
        correct_values = [0.356579, 0.560779, 0.446110, 0.179993, 0]
        self.func_test(distance.jaccard, correct_values) 

    def test_jeffreys(self):
        correct_values = [0.514598, 13.165392, 9.149020, 0.219522, 0]
        self.func_test(distance.jeffreys, correct_values) 

    def test_jensenshannon_divergence(self):
        correct_values = [0.062220, 0.191254, 0.145167, 0.027169, 0]
        self.func_test(distance.jensenshannon_divergence, correct_values)        

    def test_jensen_difference(self):
        correct_values = [0.062220, 0.191254, 0.145167, 0.027169, 0]
        self.func_test(distance.jensen_difference, correct_values) 

    def test_k_divergence(self):
        correct_values = [0.061723, 0.146679, 0.176323, 0.0266046, 0]
        self.func_test(distance.k_divergence, correct_values)

    def test_kl_divergence(self):
        correct_values = [0.259584, 12.480004, 0.523377, 0.112098, 0]
        self.func_test(distance.kl_divergence, correct_values)

    def test_kulczynski(self):
        correct_values = [0.816901, 1.773585, 1.448276, 0.597403, 0]
        self.func_test(distance.kulczynski, correct_values)

    def test_kumarjohnson(self):
        correct_values = [1.297539, 1.549443, 1.2375098, 0.470667, 0]
        self.func_test(distance.kumarjohnson, correct_values)

    def test_lorentzian(self):
        correct_values = [0.511453, 0.797107, 0.730804, 0.414028, 0]
        self.func_test(distance.lorentzian, correct_values)

    def test_manhattan(self):
        correct_values = [0.58, 0.94, 0.84, 0.46, 0]
        self.func_test(distance.manhattan, correct_values)

    def test_marylandbridge(self):
        correct_values = [0.216404, 0.350152, 0.258621, 0.096073, 0]
        self.func_test(distance.marylandbridge, correct_values)

    def test_matusita(self):
        correct_values = [0.355673, 0.704327, 0.607504, 0.233683, 0]
        self.func_test(distance.matusita, correct_values)

    def test_max_symmetric_chisq(self):
        correct_values = [0.604068, 1.349947, 1.0134, 0.243107, 0]
        self.func_test(distance.max_symmetric_chisq, correct_values)

    def test_minkowski(self):
        correct_values = [0.403237, 0.602163, 0.517494, 0.325269, 0]
        self.func_test(distance.minkowski, correct_values)

    def test_motyka(self):
        correct_values = [0.645, 0.7349999, 0.71, 0.615, 0.5]
        self.func_test(distance.motyka, correct_values)

    def test_neyman_chisq(self):
        correct_values = [0.571214, 1.349947, 0.350859, 0.213737, 0]
        self.func_test(distance.neyman_chisq, correct_values)

    def test_nonintersection(self):
        correct_values = [0.29, 0.47, 0.42, 0.23, 0]
        self.func_test(distance.nonintersection, correct_values)

    def test_pearson_chisq(self):
        correct_values = [0.604068, 0.355905, 1.0134, 0.243107, 0]
        self.func_test(distance.pearson_chisq, correct_values)

    def test_penroseshape(self):
        correct_values = [0.403237, 0.602163, 0.517494, 0.325269, 0]
        self.func_test(distance.penroseshape, correct_values)

    def test_soergel(self):
        correct_values = [0.449612, 0.639456, 0.591549, 0.373984, 0]
        self.func_test(distance.soergel, correct_values) 

    def test_squared_euclidean(self):
        correct_values = [0.1626, 0.3626, 0.2678, 0.1058, 0]
        self.func_test(distance.squared_euclidean, correct_values)

    def test_squared_chisq(self):
        correct_values = [0.241105, 0.614569, 0.476558, 0.107619, 0]
        self.func_test(distance.squared_chisq, correct_values)

    def test_squaredchord(self):
        correct_values = [0.126503, 0.496077, 0.369061, 0.054608, 0]
        self.func_test(distance.squaredchord, correct_values)

    def test_taneja(self):
        correct_values = [0.066429, 3.100094, 2.142089, 0.027711, 0]
        self.func_test(distance.taneja, correct_values)

    def test_tanimoto(self):
        correct_values = [0.449612, 0.639456, 0.591549, 0.373984, 0]
        self.func_test(distance.tanimoto, correct_values)

    def test_topsoe(self):
        correct_values = [0.124441, 0.382508, 0.290333, 0.054338, 0]
        self.func_test(distance.topsoe, correct_values)

    def test_vicis_wave_hedges(self):
        correct_values = [2.967758, 2.410145, 2.195152, 1.229861, 0]
        self.func_test(distance.vicis_wave_hedges, correct_values)

    def test_vicis_symmetric_chisq(self):
        correct_values = [4.327759, 4.310248, 3.087781, 0.777836, 0]
        self.func_test(distance.vicis_symmetric_chisq, correct_values)

    def test_wave_hedges(self):
        correct_values = [1.218999, 1.939721, 1.966866, 0.756417, 0]
        self.func_test(distance.wave_hedges, correct_values)

    def test_rel_acc_manhattan_chebyshev(self):
        for u, v in self.vectors:
            self.assertEqual( 
                distance.acc(u, v),
                (distance.manhattan(u, v) + distance.chebyshev(u, v)) / 2
            )

    def test_rel_braycurtis_manhattan(self):
        for u, v in self.vectors:
            self.assertEqual( 
                distance.braycurtis(u, v),
                distance.manhattan(u, v) / 2
            )

    def test_rel_clark_divergence(self):
        for u, v in self.vectors:
            self.assertEqual( 
                distance.clark(u, v), np.sqrt(distance.divergence(u, v) / 2)
            )

    def test_rel_euclidean_square_euclidean(self):
        for u, v in self.vectors:
            self.assertAlmostEqual( 
                distance.euclidean(u, v)**2,
                distance.squared_euclidean(u, v), places=6
            )

    def test_rel_google_manhattan(self):
        for u, v in self.vectors:
            self.assertAlmostEqual( 
                distance.google(u, v), distance.manhattan(u, v) / 2, places=6
            )

    def test_rel_gower_manhattan(self):
        for u, v in self.vectors:
            self.assertAlmostEqual( 
                distance.gower(u, v), distance.manhattan(u, v) / len(u), places=6
            )

    def test_rel_jensenshannon_divergence_topsoe(self):
        for u, v in self.vectors:
            self.assertAlmostEqual( 
                distance.jensenshannon_divergence(u, v),
                distance.topsoe(u, v) / 2, places=6
            )

    def test_rel_matusita_squaredchor(self):
        for u, v in self.vectors:
            self.assertAlmostEqual( 
                distance.matusita(u, v),
                np.sqrt(distance.squaredchord(u, v)), places=6
            )

    def test_rel_nonintersection_manhattan(self):
        for u, v in self.vectors:
            self.assertAlmostEqual( 
                distance.nonintersection(u, v),
                distance.manhattan(u, v) / 2, places=6
            )

    def test_rel_soergel_tanimoto(self):
        for u, v in self.vectors:
            self.assertAlmostEqual( 
                distance.soergel(u, v),
                distance.tanimoto(u, v), places=6
            )

    def test_rel_squaredchord_matusita(self):
        for u, v in self.vectors:
            self.assertAlmostEqual( 
                distance.squaredchord(u, v),
                distance.matusita(u, v)**2, places=6
            )

unittest.main()